#include <windows.h>
#include <shlwapi.h>

#pragma comment(lib, "shlwapi.lib")

class CFileReader {
public:
	CFileReader(LPCSTR lpFileName) {
		m_dwReadLength = 0;
		m_hFile = CreateFile(lpFileName, GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
		if(m_hFile == INVALID_HANDLE_VALUE) {
			m_hFile = NULL;
		}
	}
	virtual ~CFileReader() {
		if(m_hFile != NULL) {
			CloseHandle(m_hFile);
		}
	}

public:
	virtual bool Read(LPVOID lpBuffer, DWORD nNumberOfBytesToRead) {
		return ReadFile(m_hFile, lpBuffer, nNumberOfBytesToRead, &m_dwReadLength, NULL) != 0;
	}
	virtual DWORD GetReadLength() { 
		return m_dwReadLength;
	}
	virtual LONG Seek(LONG lDistanceToMove, DWORD dwMoveMethod) {
		return SetFilePointer(m_hFile, lDistanceToMove, NULL, dwMoveMethod);
	}
	virtual DWORD GetLength() { 
		return GetFileSize(m_hFile, NULL);
	}

public:
	HANDLE m_hFile;
	DWORD  m_dwReadLength;
};

struct CFakeSFLReader {
private:
	struct FakeSFL {
		DWORD cue;
		DWORD adtl;
		DWORD _padding[2];
		DWORD dwIntroSamples;
		DWORD dwLoopSamples;
	};

public:
	CFakeSFLReader(DWORD dwIntroSamples, DWORD dwLoopSamples) :
		m_dwIntroSamples(dwIntroSamples), m_dwLoopSamples(dwLoopSamples) {}
	virtual ~CFakeSFLReader() {}

public:
	virtual bool Read(LPVOID lpBuffer, DWORD nNumberOfBytesToRead) {
		FakeSFL *fake = (FakeSFL *)lpBuffer;
		fake->cue  = MAKEFOURCC('c','u','e',' ');
		fake->adtl = MAKEFOURCC('a','d','t','l');
		fake->_padding[0] = fake->_padding[1] = 0;
		fake->dwIntroSamples = m_dwIntroSamples;
		fake->dwLoopSamples  = m_dwLoopSamples;
		return true;
	}
	virtual DWORD GetReadLength() { 
		return sizeof(FakeSFL);
	}
	virtual LONG Seek(LONG lDistanceToMove, DWORD dwMoveMethod) {
		return 0;
	}
	virtual DWORD GetLength() {
		return sizeof(FakeSFL);
	}

private:
	DWORD m_dwIntroSamples;
	DWORD m_dwLoopSamples;
};

struct CFileReaderPtr {
	CFileReader *m_pReader;
};

static struct {
	LPCSTR lpName;
	DWORD dwIntroSamples;
	DWORD dwLoopSamples;
	LPSTR lpCustomPath;
} s_entries[] = {
	"sr", 0, 0, NULL,
	"op", 0, 0, NULL,
	"st00", 0, 0, NULL,
	"st01", 0, 0, NULL,
	"st02", 0, 0, NULL,
	"st03", 0, 0, NULL,
	"st04", 0, 0, NULL,
	"st05", 0, 0, NULL,
	"st06", 0, 0, NULL,
	"st07", 0, 0, NULL,
	"st08", 0, 0, NULL,
	"st09", 0, 0, NULL,
	"st10", 0, 0, NULL,
	"st11", 0, 0, NULL,
	"st12", 0, 0, NULL,
	"st13", 0, 0, NULL,
	"st14", 0, 0, NULL,
	"st15", 0, 0, NULL,
	"st16", 0, 0, NULL,
	"st17", 0, 0, NULL,
	"st18", 0, 0, NULL,
	"st19", 0, 0, NULL,
	"st20", 0, 0, NULL,
	"st21", 0, 0, NULL,
	"st22", 0, 0, NULL,
	"ta00", 0, 0, NULL,
	"ta01", 0, 0, NULL,
	"ta02", 0, 0, NULL,
	"ta03", 0, 0, NULL,
	"ta04", 0, 0, NULL,
	"ta05", 0, 0, NULL,
	"ta06", 0, 0, NULL,
	"ta07", 0, 0, NULL,
	"ta08", 0, 0, NULL,
};

static bool (__fastcall *GetPackagedBGM)(CFileReaderPtr *p, int, LPCSTR lpEntryName);
static bool (__fastcall *GetPackagedSFL)(CFileReaderPtr *p, int, LPCSTR lpEntryName);

#ifndef _DEBUG
static HANDLE s_hHeap;

void * operator new(size_t size)
{
	return HeapAlloc(s_hHeap, 0, size);
}

void operator delete(void *p)
{
	HeapFree(s_hHeap, 0, p);
}

#endif

int GetEntryIndex(LPCSTR lpEntryName)
{
	for(int i = 0; i < _countof(s_entries); ++i) {
		if(s_entries[i].lpCustomPath[0] != '\0' &&
		   StrStr(lpEntryName, s_entries[i].lpName)) {
			return i;
		}
	}
	return -1;
}

bool __fastcall OnGetPackagedBGM(CFileReaderPtr *p, int, LPCSTR lpEntryName)
{
	int nEntryIndex = GetEntryIndex(lpEntryName);
	if(nEntryIndex < 0) {
		return GetPackagedBGM(p, 0, lpEntryName);
	}

	if(p->m_pReader != NULL) {
		delete p->m_pReader;
		p->m_pReader = NULL;
	}
	p->m_pReader = new CFileReader(s_entries[nEntryIndex].lpCustomPath);

	if(p->m_pReader->m_hFile == NULL) {
		return GetPackagedBGM(p, 0, s_entries[nEntryIndex].lpCustomPath);
	} else {
		return true;
	}
}

bool __fastcall OnGetPackagedSFL(CFileReaderPtr *p, int, LPCSTR lpEntryName)
{
	int nEntryIndex = GetEntryIndex(lpEntryName);
	if(nEntryIndex < 0) {
		return GetPackagedSFL(p, 0, lpEntryName);
	}

	if(p->m_pReader != NULL) {
		delete p->m_pReader;
		p->m_pReader = NULL;
	}
	p->m_pReader = (CFileReader *)
		new CFakeSFLReader(
			s_entries[nEntryIndex].dwIntroSamples,
			s_entries[nEntryIndex].dwLoopSamples);
	return true;
}

BOOL HookHisouten(HMODULE hModule)
{
#ifndef _DEBUG
	s_hHeap = ::HeapCreate(0, 1024 * _countof(s_entries), 0);
#endif
	char szProfilePath[1024 + MAX_PATH];
	GetModuleFileName(hModule, szProfilePath, 1024);
	PathRemoveFileSpec(szProfilePath);
	PathAppend(szProfilePath, "BGMChanger.ini");

	HANDLE hHeap = GetProcessHeap();
	for(int i = 0; i < _countof(s_entries); ++i) {
		s_entries[i].lpCustomPath = (LPSTR)HeapAlloc(hHeap, 0, 1024);
		GetPrivateProfileString(s_entries[i].lpName, "Path", "",
			s_entries[i].lpCustomPath, 1024, szProfilePath);
		s_entries[i].dwIntroSamples =
			GetPrivateProfileInt(s_entries[i].lpName, "Intro", 0, szProfilePath);
		s_entries[i].dwLoopSamples =
			GetPrivateProfileInt(s_entries[i].lpName, "Loop", 0, szProfilePath);
	}

	DWORD dwOldProtect;
	::VirtualProtect((PVOID)0x401000, 0x2B6000, PAGE_EXECUTE_WRITECOPY, &dwOldProtect);
	*(DWORD*)&GetPackagedBGM = *(DWORD*)0x418592 + (0x418591 + 5);
	*(DWORD*)0x418592 = (DWORD)OnGetPackagedBGM - (0x418591 + 5);
	*(DWORD*)&GetPackagedSFL = *(DWORD*)0x418676 + (0x418675 + 5);
	*(DWORD*)0x418676 = (DWORD)OnGetPackagedSFL - (0x418675 + 5);
	::VirtualProtect((PVOID)0x401000, 0x2B6000, dwOldProtect, &dwOldProtect);

	return TRUE;
}

extern "C"
int APIENTRY DllMain(HMODULE hModule, DWORD fdwReason, LPVOID lpReserved)
{
	switch(fdwReason) {
	case DLL_PROCESS_ATTACH:
		DisableThreadLibraryCalls(hModule);
		return HookHisouten(hModule);
	}
	return TRUE;
}
