#include <windows.h>
#include <shlwapi.h>

#pragma comment (lib, "shlwapi.lib")

static char s_profilePath[1024 + MAX_PATH];

static bool s_sizeEnabled;
static int s_sizeWidth;
static int s_sizeHeight;
static int s_sizeWindowWidth;
static int s_sizeWindowHeight;

static bool s_posEnabled;
static int s_posX;
static int s_posY;
static WNDPROC s_origWndProc;

void MyAdjustClientRect(RECT *prect)
{
	int cx  = ::GetSystemMetrics(SM_CXSIZEFRAME);
	int cy1 = ::GetSystemMetrics(SM_CYSIZEFRAME);
	int cy2 = cy1 + ::GetSystemMetrics(SM_CYCAPTION);
	prect->left += cx; prect->right -= cx;
	prect->top += cy2; prect->bottom -= cy1;
}

void MyAdjustWindowRect(RECT *prect)
{
	int cx  = ::GetSystemMetrics(SM_CXSIZEFRAME);
	int cy1 = ::GetSystemMetrics(SM_CYSIZEFRAME);
	int cy2 = cy1 + ::GetSystemMetrics(SM_CYCAPTION);
	prect->left -= cx; prect->right += cx;
	prect->top -= cy2; prect->bottom += cy1;
}

LRESULT CALLBACK OnWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	if(s_sizeEnabled && uMsg == WM_SIZING) {
		RECT &rct = *(RECT*)lParam;
		MyAdjustClientRect(&rct);
		switch(wParam) {
		case WMSZ_LEFT:
		case WMSZ_RIGHT:
			rct.bottom = MulDiv(rct.right - rct.left, 3, 4) + rct.top;
			break;
		case WMSZ_TOPLEFT:
		case WMSZ_BOTTOMLEFT:
			rct.left = - MulDiv(rct.bottom - rct.top, 4, 3) + rct.right;
			break;
		default:
			rct.right = MulDiv(rct.bottom - rct.top, 4, 3) + rct.left;
			break;
		}
		s_sizeWidth  = rct.right - rct.left;
		s_sizeHeight = rct.bottom - rct.top;
		MyAdjustWindowRect(&rct);

		s_sizeWindowWidth  = rct.right - rct.left;
		s_sizeWindowHeight = rct.bottom - rct.top;
		return TRUE;
	} else if(s_posEnabled && uMsg == WM_MOVE) {
		RECT rct = { (short)LOWORD(lParam), (short)HIWORD(lParam), 0, 0 };
		MyAdjustWindowRect(&rct);
		s_posX = rct.left;
		s_posY = rct.top;
	}

	return s_origWndProc(hWnd, uMsg, wParam, lParam);
}

BOOL (__stdcall *s_SetWindowPos)(
  HWND hWnd,
  HWND hWndInsertAfter,
  int X,
  int Y,
  int cx,
  int cy,
  UINT uFlags
);


BOOL __stdcall OnSetWindowPos(
  HWND hWnd,
  HWND hWndInsertAfter,
  int X,
  int Y,
  int cx,
  int cy,
  UINT uFlags
)
{
	if(hWndInsertAfter != HWND_TOP) {
		if(s_sizeEnabled) {
			uFlags &= ~SWP_NOSIZE;
			cx = s_sizeWindowWidth;
			cy = s_sizeWindowHeight;
		}
		if(s_posEnabled) {
			if(s_posX != CW_USEDEFAULT) X = s_posX;
			if(s_posY != CW_USEDEFAULT) Y = s_posY;
		}
	}
	return s_SetWindowPos(hWnd, hWndInsertAfter, X, Y, cx, cy, uFlags);
}


HWND (__stdcall *s_CreateWindowExA)(
  DWORD dwExStyle,
  LPCSTR lpClassName,
  LPCSTR lpWindowName,
  DWORD dwStyle,
  int x,
  int y,
  int nWidth,
  int nHeight,
  HWND hWndParent,
  HMENU hMenu,
  HINSTANCE hInstance,
  LPVOID lpParam
);

HWND __stdcall OnCreateWindowExA(
  DWORD dwExStyle,
  LPCSTR lpClassName,
  LPCSTR lpWindowName,
  DWORD dwStyle,
  int x,
  int y,
  int nWidth,
  int nHeight,
  HWND hWndParent,
  HMENU hMenu,
  HINSTANCE hInstance,
  LPVOID lpParam
)
{
	if(s_sizeEnabled) {
		dwStyle = WS_OVERLAPPED | WS_GROUP | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME;
		RECT rct = { 0, 0, s_sizeWidth, s_sizeHeight };
		::AdjustWindowRectEx(&rct, dwStyle, FALSE, dwExStyle);
		nWidth  = s_sizeWindowWidth  = rct.right  - rct.left;
		nHeight = s_sizeWindowHeight = rct.bottom - rct.top;
	}

	if(s_posEnabled) {
		x = s_posX;
		y = s_posY;
	}

	return s_CreateWindowExA(dwExStyle, lpClassName, lpWindowName, dwStyle, x, y, nWidth, nHeight, hWndParent, hMenu, hInstance, lpParam);
}

BOOL __stdcall OnGetWindowInfo(HWND hwnd, PWINDOWINFO pwi)
{
	pwi->rcClient.left = pwi->rcClient.top = 0;
	pwi->rcClient.right = 640; pwi->rcClient.bottom = 480;
	return TRUE;
}

void (__fastcall *s_MainLoop)(void *);

void __fastcall OnMainLoop(void *This)
{
	s_MainLoop(This);

	if(s_sizeEnabled) {
		char buff[16];
		::wsprintf(buff, "%d", s_sizeWidth);
		::WritePrivateProfileString("Size", "Width", buff, s_profilePath); 
		::wsprintf(buff, "%d", s_sizeHeight);
		::WritePrivateProfileString("Size", "Height", buff, s_profilePath); 
	}
	if(s_posEnabled) {
		char buff[16];
		::wsprintf(buff, "%d", s_posX);
		::WritePrivateProfileString("Position", "X", buff, s_profilePath); 
		::wsprintf(buff, "%d", s_posY);
		::WritePrivateProfileString("Position", "Y", buff, s_profilePath); 
	}
}


void LoadSettings(LPCSTR profilePath)
{
	s_sizeEnabled = ::GetPrivateProfileInt("Size", "Enabled", 0, profilePath) != 0;
	if(s_sizeEnabled) {
		s_sizeWidth = ::GetPrivateProfileInt("Size", "Width", 640, profilePath);
		s_sizeHeight = ::MulDiv(s_sizeWidth, 3, 4);
	}
	s_posEnabled = GetPrivateProfileInt("Position", "Enabled", 0, profilePath) != 0;
	if(s_posEnabled) {
		s_posX = ::GetPrivateProfileInt("Position", "X", CW_USEDEFAULT, profilePath);
		s_posY = ::GetPrivateProfileInt("Position", "Y", CW_USEDEFAULT, profilePath);
	}
}

BOOL Hook(HMODULE hModule)
{
	::GetModuleFileName(hModule, s_profilePath, 1024);
	::PathRemoveFileSpec(s_profilePath);
	::PathAppend(s_profilePath, "WindowResizer.ini");
	::LoadSettings(s_profilePath);

	DWORD old;
	::VirtualProtect((PVOID)0x401000, 0x2B6000, PAGE_EXECUTE_WRITECOPY, &old);
	*(DWORD*)&s_origWndProc = *(DWORD*)0x66F4F2;
	*(DWORD*)0x66F4F2 = (DWORD)OnWindowProc;
	*(DWORD*)&s_MainLoop = *(DWORD*)0x40884D + (0x40884C + 5);
	*(DWORD*)0x40884D = (DWORD)OnMainLoop - (0x40884C + 5);
	::VirtualProtect((PVOID)0x401000, 0x2B6000, old, &old);

	::VirtualProtect((PVOID)0x6B7000, 0x2D4, PAGE_EXECUTE_WRITECOPY, &old);
	*(DWORD*)&s_CreateWindowExA = *(DWORD*)0x6B7228;
	*(DWORD*)0x6B7228 = (DWORD)OnCreateWindowExA;
	*(DWORD*)&s_SetWindowPos = *(DWORD*)0x6B7258;
	*(DWORD*)0x6B7258 = (DWORD)OnSetWindowPos;
	*(DWORD*)0x6B7254 = (DWORD)OnGetWindowInfo;
	::VirtualProtect((PVOID)0x6B7000, 0x2D4, old, &old);

	return TRUE;
}

extern "C"
int APIENTRY DllMain(HMODULE hModule, DWORD fdwReason, LPVOID lpReserved)
{
	return (fdwReason == DLL_PROCESS_ATTACH) ? Hook(hModule) : TRUE;
}
